# core client classes
from .core import (
    namespace,
    get_namespace,
    default_namespace,
    metadata,
    get_metadata,
    default_metadata,
    inspect_spin,
    Metaflow,
    Flow,
    Run,
    Step,
    Task,
    DataArtifact,
)



from typing import Any, Dict, Optional, Union

from metaflow.metaflow_config import DEFAULT_SECRETS_ROLE
from metaflow.exception import MetaflowException
from metaflow.plugins.secrets.secrets_spec import SecretSpec
from metaflow.plugins.secrets.utils import get_secrets_backend_provider


def get_secret(
    source: Union[str, Dict[str, Any]], role: Optional[str] = None
) -> Dict[str, str]:
    """
    Get secret from source

    Parameters
    ----------
    source : Union[str, Dict[str, Any]]
        Secret spec, defining how the secret is to be retrieved
    role : str, optional
        Role to use for fetching secrets
    """
    if role is None:
        role = DEFAULT_SECRETS_ROLE

    secret_spec = None

    if isinstance(source, str):
        secret_spec = SecretSpec.secret_spec_from_str(source, role=role)
    elif isinstance(source, dict):
        secret_spec = SecretSpec.secret_spec_from_dict(source, role=role)
    else:
        raise MetaflowException(
            "get_secrets sources items must be either a string or a dict"
        )

    secrets_backend_provider = get_secrets_backend_provider(
        secret_spec.secrets_backend_type
    )
    try:
        dict_for_secret = secrets_backend_provider.get_secret_as_dict(
            secret_spec.secret_id,
            options=secret_spec.options,
            role=secret_spec.role,
        )
        return dict_for_secret
    except Exception as e:
        raise MetaflowException(
            "Failed to retrieve secret '%s': %s" % (secret_spec.secret_id, e)
        )



import os
import shutil
import subprocess
import sys
import time

from metaflow.util import which
from metaflow.meta_files import read_info_file
from metaflow.metaflow_config import get_pinned_conda_libs
from metaflow.packaging_sys import MetaflowCodeContent, ContentType
from urllib.request import Request, urlopen
from urllib.error import URLError

# TODO: support version/platform/architecture selection.
UV_URL = "https://github.com/astral-sh/uv/releases/download/0.6.11/uv-x86_64-unknown-linux-gnu.tar.gz"

if __name__ == "__main__":

    def run_cmd(cmd, stdin_str=None):
        result = subprocess.run(
            cmd,
            shell=True,
            input=stdin_str,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        if result.returncode != 0:
            print(f"Bootstrap failed while executing: {cmd}")
            print("Stdout:", result.stdout)
            print("Stderr:", result.stderr)
            sys.exit(1)

    def install_uv():
        import tarfile

        uv_install_path = os.path.join(os.getcwd(), "uv_install")
        if which("uv"):
            return

        print("Installing uv...")

        # Prepare directory once
        os.makedirs(uv_install_path, exist_ok=True)

        # Download and decompress in one go
        headers = {
            "Accept-Encoding": "gzip, deflate, br",
            "Connection": "keep-alive",
            "User-Agent": "python-urllib",
        }

        def _tar_filter(member: tarfile.TarInfo, path):
            if os.path.basename(member.name) != "uv":
                return None  # skip
            member.path = os.path.basename(member.path)
            return member

        max_retries = 3
        for attempt in range(max_retries):
            try:
                req = Request(UV_URL, headers=headers)
                with urlopen(req) as response:
                    with tarfile.open(fileobj=response, mode="r:gz") as tar:
                        tar.extractall(uv_install_path, filter=_tar_filter)
                break
            except (URLError, IOError) as e:
                if attempt == max_retries - 1:
                    raise Exception(
                        f"Failed to download UV after {max_retries} attempts: {e}"
                    )
                time.sleep(2**attempt)

        # Update PATH only once at the end
        os.environ["PATH"] += os.pathsep + uv_install_path

    def get_dependencies(datastore_type):
        # return required dependencies for Metaflow that must be added to the UV environment.
        pinned = get_pinned_conda_libs(None, datastore_type)

        # return only dependency names instead of pinned versions
        return pinned.keys()

    def skip_metaflow_dependencies():
        skip_pkgs = ["metaflow"]
        info = read_info_file()
        if info is not None:
            try:
                skip_pkgs.extend([ext_name for ext_name in info["ext_info"][0].keys()])
            except Exception:
                print(
                    "Failed to read INFO. Metaflow-related packages might get installed during runtime."
                )

        return skip_pkgs

    def sync_uv_project(datastore_type):
        # Move the files to the current directory so uv can find them.
        for filename in ["uv.lock", "pyproject.toml"]:
            path_to_file = MetaflowCodeContent.get_filename(
                filename, ContentType.OTHER_CONTENT
            )
            if path_to_file is None:
                raise RuntimeError(f"Could not find {filename} in the package.")
            shutil.move(path_to_file, os.path.join(os.getcwd(), filename))

        print("Syncing uv project...")
        dependencies = " ".join(get_dependencies(datastore_type))
        skip_pkgs = " ".join(
            [f"--no-install-package {dep}" for dep in skip_metaflow_dependencies()]
        )
        cmd = f"""set -e;
            uv sync --frozen --no-dev {skip_pkgs};
            uv pip install {dependencies} --strict
            """
        run_cmd(cmd)

    if len(sys.argv) != 2:
        print("Usage: bootstrap.py <datastore_type>")
        sys.exit(1)

    try:
        datastore_type = sys.argv[1]
        install_uv()
        sync_uv_project(datastore_type)
    except Exception as e:
        print(f"Error: {str(e)}", file=sys.stderr)
        sys.exit(1)



from abc import ABC, abstractmethod
from io import BytesIO
from typing import Any, IO, List, Optional, Union


class PackagingBackend(ABC):
    _mappings = {}
    type = "none"

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls.type in cls._mappings:
            raise ValueError(f"PackagingBackend {cls.type} already exists")
        cls._mappings[cls.type] = cls

    @classmethod
    def get_backend(cls, name: str) -> "PackagingBackend":
        if name not in cls._mappings:
            raise ValueError(f"PackagingBackend {name} not found")
        return cls._mappings[name]

    @classmethod
    def backend_type(cls) -> str:
        return cls.type

    @classmethod
    @abstractmethod
    def get_extract_commands(cls, archive_name: str, dest_dir: str) -> List[str]:
        pass

    def __init__(self):
        self._archive = None

    @abstractmethod
    def create(self) -> "PackagingBackend":
        pass

    @abstractmethod
    def add_file(self, filename: str, arcname: Optional[str] = None):
        pass

    @abstractmethod
    def add_data(self, data: BytesIO, arcname: str):
        pass

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def get_blob(self) -> Optional[Union[bytes, bytearray]]:
        pass

    @classmethod
    @abstractmethod
    def cls_open(cls, content: IO[bytes]) -> Any:
        """Open the archive from the given content."""
        pass

    @classmethod
    @abstractmethod
    def cls_member_name(cls, member: Union[Any, str]) -> str:
        """
        Returns the name of the member as a string.
        This is used to ensure consistent naming across different archive formats.
        """
        pass

    @classmethod
    @abstractmethod
    def cls_has_member(cls, archive: Any, name: str) -> bool:
        pass

    @classmethod
    @abstractmethod
    def cls_get_member(cls, archive: Any, name: str) -> Optional[bytes]:
        pass

    @classmethod
    @abstractmethod
    def cls_extract_members(
        cls,
        archive: Any,
        members: Optional[List[Any]] = None,
        dest_dir: str = ".",
    ) -> None:
        pass

    @classmethod
    @abstractmethod
    def cls_list_names(cls, archive: Any) -> Optional[List[str]]:
        pass

    @classmethod
    @abstractmethod
    def cls_list_members(cls, archive: Any) -> Optional[List[Any]]:
        """List all members in the archive."""
        pass

    def has_member(self, name: str) -> bool:
        if self._archive:
            return self.cls_has_member(self._archive, name)
        raise ValueError("Cannot check for member in an uncreated archive")

    def get_member(self, name: str) -> Optional[bytes]:
        if self._archive:
            return self.cls_get_member(self._archive, name)
        raise ValueError("Cannot get member from an uncreated archive")

    def extract_members(
        self, members: Optional[List[Any]] = None, dest_dir: str = "."
    ) -> None:
        if self._archive:
            self.cls_extract_members(self._archive, members, dest_dir)
        else:
            raise ValueError("Cannot extract from an uncreated archive")

    def list_names(self) -> Optional[List[str]]:
        if self._archive:
            return self.cls_list_names(self._archive)
        raise ValueError("Cannot list names from an uncreated archive")

    def __enter__(self):
        self.create()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()



import functools
import json
import os
import re
import shutil
import subprocess
import tempfile
import time

from metaflow.debug import debug
from metaflow.exception import MetaflowException
from metaflow.util import which

from .utils import MICROMAMBA_MIRROR_URL, MICROMAMBA_URL, conda_platform
from threading import Lock


class MicromambaException(MetaflowException):
    headline = "Micromamba ran into an error while setting up environment"

    def __init__(self, error):
        if isinstance(error, (list,)):
            error = "\n".join(error)
        msg = "{error}".format(error=error)
        super(MicromambaException, self).__init__(msg)


GLIBC_VERSION = os.environ.get("CONDA_OVERRIDE_GLIBC", "2.38")

_double_equal_match = re.compile("==(?=[<=>!~])")


class Micromamba(object):
    def __init__(self, logger=None, force_rebuild=False):
        # micromamba is a tiny version of the mamba package manager and comes with
        # metaflow specific performance enhancements.

        # METAFLOW_HOME might not be writable but METAFLOW_TOKEN_HOME might be.
        if os.environ.get("METAFLOW_TOKEN_HOME"):
            _home = os.environ.get("METAFLOW_TOKEN_HOME")
        else:
            _home = os.environ.get("METAFLOW_HOME", "~/.metaflowconfig")
        self._path_to_hidden_micromamba = os.path.join(
            os.path.expanduser(_home),
            "micromamba",
        )

        if logger:
            self.logger = logger
        else:
            self.logger = lambda *args, **kwargs: None  # No-op logger if not provided

        self._bin = (
            which(os.environ.get("METAFLOW_PATH_TO_MICROMAMBA") or "micromamba")
            or which("./micromamba")  # to support remote execution
            or which("./bin/micromamba")
            or which(os.path.join(self._path_to_hidden_micromamba, "bin/micromamba"))
        )

        # We keep a mutex as environments are resolved in parallel,
        # which causes a race condition in case micromamba needs to be installed first.
        self.install_mutex = Lock()

        self.force_rebuild = force_rebuild

    @property
    def bin(self) -> str:
        "Defer installing Micromamba until when the binary path is actually requested"
        if self._bin is not None:
            return self._bin
        with self.install_mutex:
            # another check as micromamba might have been installed when the mutex is released.
            if self._bin is not None:
                return self._bin

            # Install Micromamba on the fly.
            # TODO: Make this optional at some point.
            debug.conda_exec("No Micromamba binary found. Installing micromamba")
            _install_micromamba(self._path_to_hidden_micromamba)
            self._bin = which(
                os.path.join(self._path_to_hidden_micromamba, "bin/micromamba")
            )

            if self._bin is None:
                msg = "No installation for *Micromamba* found.\n"
                msg += "Visit https://mamba.readthedocs.io/en/latest/micromamba-installation.html for installation instructions."
                raise MetaflowException(msg)

        return self._bin

    def solve(self, id_, packages, python, platform):
        # Performance enhancements
        # 1. Using zstd compressed repodata index files drops the index download time
        #    by a factor of 10x - conda-forge/noarch/repodata.json has a
        #    mean download time of 3.705s ± 2.283s vs 385.1ms ± 57.9ms for
        #    conda-forge/noarch/repodata.json.zst. Thankfully, now micromamba pulls
        #    zstd compressed files by default - https://github.com/conda-forge/conda-forge.github.io/issues/1835
        # 2. Tweak repodata ttl to pull either only once a day or if a solve fails -
        #    this ensures that repodata is not pulled everytime it becomes dirty by
        #    default. This can result in an environment that has stale transitive
        #    dependencies but still correct. (--repodata-ttl 86400 --retry-clean-cache)
        # 3. Introduce pip as python dependency to resolve pip packages within conda
        #    environment
        # 4. Multiple solves can progress at the same time while relying on the same
        #    index
        debug.conda_exec("Solving packages for conda environment %s" % id_)
        with tempfile.TemporaryDirectory() as tmp_dir:
            env = {
                "MAMBA_ADD_PIP_AS_PYTHON_DEPENDENCY": "true",
                "CONDA_SUBDIR": platform,
                # "CONDA_UNSATISFIABLE_HINTS_CHECK_DEPTH": "0" # https://github.com/conda/conda/issues/9862
                # Add a default glibc version for linux-64 environments (ignored for other platforms)
                # TODO: Make the version configurable
                "CONDA_OVERRIDE_GLIBC": GLIBC_VERSION,
            }
            cmd = [
                "create",
                "--yes",
                "--quiet",
                "--dry-run",
                "--no-extra-safety-checks",
                "--repodata-ttl=86400",
                "--safety-checks=disabled",
                "--retry-clean-cache",
                "--prefix=%s/prefix" % tmp_dir,
            ]
            # Introduce conda-forge as a default channel
            for channel in self.info()["channels"] or ["conda-forge"]:
                cmd.append("--channel=%s" % channel)

            for package, version in packages.items():
                version_string = "%s==%s" % (package, version)
                cmd.append(_double_equal_match.sub("", version_string))
            if python:
                cmd.append("python==%s" % python)
            # TODO: Ensure a human readable message is returned when the environment
            #       can't be resolved for any and all reasons.
            solved_packages = [
                {k: v for k, v in item.items() if k in ["url"]}
                for item in self._call(cmd, env)["actions"]["LINK"]
            ]
            return solved_packages

    def download(self, id_, packages, python, platform):
        # Unfortunately all the packages need to be catalogued in package cache
        # because of which this function can't be parallelized

        # Micromamba is painfully slow in determining if many packages are infact
        # already cached. As a perf heuristic, we check if the environment already
        # exists to short circuit package downloads.

        prefix = "{env_dirs}/{keyword}/{platform}/{id}".format(
            env_dirs=self.info()["envs_dirs"][0],
            platform=platform,
            keyword="metaflow",  # indicates metaflow generated environment
            id=id_,
        )
        # If we are forcing a rebuild of the environment, we make sure to remove existing files beforehand.
        # This is to ensure that no irrelevant packages get bundled relative to the resolved environment.
        # NOTE: download always happens before create, so we want to do the cleanup here instead.
        if self.force_rebuild:
            shutil.rmtree(self.path_to_environment(id_, platform), ignore_errors=True)

        # cheap check
        if os.path.exists(f"{prefix}/fake.done"):
            return

        # somewhat expensive check
        if self.path_to_environment(id_, platform):
            return

        debug.conda_exec("Downloading packages for conda environment %s" % id_)
        with tempfile.TemporaryDirectory() as tmp_dir:
            env = {
                "CONDA_SUBDIR": platform,
                "CONDA_OVERRIDE_GLIBC": GLIBC_VERSION,
            }
            cmd = [
                "create",
                "--yes",
                "--no-deps",
                "--download-only",
                "--safety-checks=disabled",
                "--no-extra-safety-checks",
                "--repodata-ttl=86400",
                "--prefix=%s/prefix" % tmp_dir,
                "--quiet",
            ]
            for package in packages:
                cmd.append("{url}".format(**package))

            self._call(cmd, env)
            # Perf optimization to skip cross-platform downloads.
            if platform != self.platform():
                os.makedirs(prefix, exist_ok=True) or open(
                    f"{prefix}/fake.done", "w"
                ).close()
            return

    def create(self, id_, packages, python, platform):
        # create environment only if the platform matches system platform
        if platform != self.platform() or self.path_to_environment(id_, platform):
            return

        debug.conda_exec("Creating local Conda environment %s" % id_)
        prefix = "{env_dirs}/{keyword}/{platform}/{id}".format(
            env_dirs=self.info()["envs_dirs"][0],
            platform=platform,
            keyword="metaflow",  # indicates metaflow generated environment
            id=id_,
        )

        env = {
            # use hardlinks when possible, otherwise copy files
            # disabled for now since it adds to environment creation latencies
            "CONDA_ALLOW_SOFTLINKS": "0",
            "CONDA_OVERRIDE_GLIBC": GLIBC_VERSION,
        }
        cmd = [
            "create",
            "--yes",
            "--no-extra-safety-checks",
            # "--offline", # not needed since micromamba will first look at the cache
            "--prefix",  # trick to ensure environments can be created in parallel
            prefix,
            "--quiet",
            "--no-deps",  # important!
        ]
        for package in packages:
            cmd.append("{url}".format(**package))
        self._call(cmd, env)

    @functools.lru_cache(maxsize=None)
    def info(self):
        return self._call(["config", "list", "-a"])

    def path_to_environment(self, id_, platform=None):
        if platform is None:
            platform = self.platform()
        suffix = "{keyword}/{platform}/{id}".format(
            platform=platform,
            keyword="metaflow",  # indicates metaflow generated environment
            id=id_,
        )
        for env in self._call(["env", "list"])["envs"]:
            # TODO: Check bin/python is available as a heuristic for well formed env
            if env.endswith(suffix):
                return env

    def metadata(self, id_, packages, python, platform):
        # this method unfortunately relies on the implementation detail for
        # conda environments and has the potential to break all of a sudden.
        packages_to_filenames = {
            package["url"]: package["url"].split("/")[-1] for package in packages
        }
        directories = self.info()["pkgs_dirs"]
        # search all package caches for packages

        file_to_path = {}
        for d in directories:
            if os.path.isdir(d):
                try:
                    with os.scandir(d) as entries:
                        for entry in entries:
                            if entry.is_file():
                                # Prefer the first occurrence if the file exists in multiple directories
                                file_to_path.setdefault(entry.name, entry.path)
                except OSError:
                    continue
        ret = {
            # set package tarball local paths to None if package tarballs are missing
            url: file_to_path.get(file)
            for url, file in packages_to_filenames.items()
        }
        return ret

    def interpreter(self, id_):
        return os.path.join(self.path_to_environment(id_), "bin/python")

    def platform(self):
        return self.info()["platform"]

    def _call(self, args, env=None):
        if env is None:
            env = {}
        try:
            result = (
                subprocess.check_output(
                    [self.bin] + args,
                    stderr=subprocess.PIPE,
                    env={
                        **os.environ,
                        # prioritize metaflow-specific env vars
                        **{k: v for k, v in env.items() if v is not None},
                        **{
                            "MAMBA_NO_BANNER": "1",
                            "MAMBA_JSON": "true",
                            # play with fire! needed for resolving cross-platform
                            # environments
                            "CONDA_SAFETY_CHECKS": "disabled",
                            # "CONDA_UNSATISFIABLE_HINTS_CHECK_DEPTH": "0",
                            # Support packages on S3
                            # "CONDA_ALLOW_NON_CHANNEL_URLS": "1",
                            "MAMBA_USE_LOCKFILES": "false",
                        },
                    },
                )
                .decode()
                .strip()
            )
            if result:
                return json.loads(result)
            return {}
        except subprocess.CalledProcessError as e:
            msg = "command '{cmd}' returned error ({code})\n{stderr}"
            try:
                output = json.loads(e.output)
                err = []
                v_pkgs = ["__cuda", "__glibc"]
                for error in output.get("solver_problems", []):
                    # raise a specific error message for virtual package related errors
                    match = next((p for p in v_pkgs if p in error), None)
                    if match is not None:
                        vpkg_name = match[2:]
                        # try to strip version from error msg which are of the format:
                        # nothing provides <__vpkg> >=2.17,<3.0.a0 needed by <pkg_name>
                        try:
                            vpkg_version = error[
                                len("nothing provides %s " % match) : error.index(
                                    " needed by"
                                )
                            ]
                        except ValueError:
                            vpkg_version = None
                        raise MicromambaException(
                            "{msg}\n\n"
                            "*Please set the environment variable CONDA_OVERRIDE_{var} to a specific version{version} of {name}.*\n\n"
                            "Here is an example of supplying environment variables through the command line\n"
                            "CONDA_OVERRIDE_{var}=<{name}-version> python flow.py <args>".format(
                                msg=msg.format(
                                    cmd=" ".join(e.cmd),
                                    code=e.returncode,
                                    output=e.output.decode(),
                                    stderr=error,
                                ),
                                var=vpkg_name.upper(),
                                version=(
                                    "" if not vpkg_version else f" ({vpkg_version})"
                                ),
                                name=vpkg_name,
                            )
                        )
                    err.append(error)
                raise MicromambaException(
                    msg.format(
                        cmd=" ".join(e.cmd),
                        code=e.returncode,
                        output=e.output.decode(),
                        stderr="\n".join(err),
                    )
                )
            except (TypeError, ValueError):
                pass
            raise MicromambaException(
                msg.format(
                    cmd=" ".join(e.cmd),
                    code=e.returncode,
                    output=e.output.decode(),
                    stderr=e.stderr.decode(),
                )
            )


def _install_micromamba(installation_location):
    # Unfortunately no 32bit binaries are available for micromamba, which ideally
    # shouldn't be much of a problem in today's world.
    platform = conda_platform()
    url = MICROMAMBA_URL.format(platform=platform, version="1.5.7")
    mirror_url = MICROMAMBA_MIRROR_URL.format(platform=platform, version="1.5.7")
    os.makedirs(installation_location, exist_ok=True)

    def _download_and_extract(url):
        max_retries = 3
        for attempt in range(max_retries):
            try:
                # https://mamba.readthedocs.io/en/latest/micromamba-installation.html#manual-installation
                # requires bzip2
                result = subprocess.Popen(
                    f"curl -Ls {url} | tar -xvj -C {installation_location} bin/micromamba",
                    shell=True,
                    stderr=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                )
                _, err = result.communicate()
                if result.returncode != 0:
                    raise MicromambaException(
                        f"Micromamba installation '{result.args}' failed:\n{err.decode()}"
                    )
            except subprocess.CalledProcessError as e:
                if attempt == max_retries - 1:
                    raise MicromambaException(
                        "Micromamba installation failed:\n{}".format(e.stderr.decode())
                    )
                time.sleep(2**attempt)

    try:
        # prioritize downloading from mirror
        _download_and_extract(mirror_url)
    except Exception:
        # download from official source as a fallback
        _download_and_extract(url)



import os

from metaflow._vendor import click


def makedirs(path):
    # This is for python2 compatibility.
    # Python3 has os.makedirs(exist_ok=True).
    try:
        os.makedirs(path)
    except OSError as x:
        if x.errno == 17:
            return
        else:
            raise


def echo_dev_null(*args, **kwargs):
    pass


def echo_always(line, **kwargs):
    click.secho(line, **kwargs)



import os
import sys
import time
import signal
import requests
import subprocess
from multiprocessing import Process
from datetime import datetime, timezone
from metaflow.sidecar import MessageTypes


class SpotTerminationMonitorSidecar(object):
    EC2_TYPE_URL = "http://169.254.169.254/latest/meta-data/instance-life-cycle"
    METADATA_URL = "http://169.254.169.254/latest/meta-data/spot/termination-time"
    TOKEN_URL = "http://169.254.169.254/latest/api/token"
    POLL_INTERVAL = 5  # seconds

    def __init__(self):
        self.is_alive = True
        self._process = None
        self._token = None
        self._token_expiry = 0

        if self._is_aws_spot_instance():
            self._process = Process(target=self._monitor_loop)
            self._process.start()

    def process_message(self, msg):
        if msg.msg_type == MessageTypes.SHUTDOWN:
            self.is_alive = False
            if self._process:
                self._process.terminate()

    @classmethod
    def get_worker(cls):
        return cls

    def _get_imds_token(self):
        current_time = time.time()
        if current_time >= self._token_expiry - 60:  # Refresh 60s before expiry
            try:
                response = requests.put(
                    url=self.TOKEN_URL,
                    headers={"X-aws-ec2-metadata-token-ttl-seconds": "300"},
                    timeout=1,
                )
                if response.status_code == 200:
                    self._token = response.text
                    self._token_expiry = current_time + 240  # Slightly less than TTL
            except requests.exceptions.RequestException:
                pass
        return self._token

    def _make_ec2_request(self, url, timeout):
        token = self._get_imds_token()
        headers = {"X-aws-ec2-metadata-token": token} if token else {}
        response = requests.get(url=url, headers=headers, timeout=timeout)
        return response

    def _is_aws_spot_instance(self):
        try:
            response = self._make_ec2_request(url=self.EC2_TYPE_URL, timeout=1)
            return response.status_code == 200 and response.text == "spot"
        except (requests.exceptions.RequestException, requests.exceptions.Timeout):
            return False

    def _monitor_loop(self):
        while self.is_alive:
            try:
                response = self._make_ec2_request(url=self.METADATA_URL, timeout=1)
                if response.status_code == 200:
                    termination_time = response.text
                    self._emit_termination_metadata(termination_time)
                    os.kill(os.getppid(), signal.SIGTERM)
                    break
            except (requests.exceptions.RequestException, requests.exceptions.Timeout):
                pass
            time.sleep(self.POLL_INTERVAL)

    def _emit_termination_metadata(self, termination_time):
        flow_filename = os.getenv("METAFLOW_FLOW_FILENAME")
        pathspec = os.getenv("MF_PATHSPEC")
        _, run_id, step_name, task_id = pathspec.split("/")
        retry_count = os.getenv("MF_ATTEMPT")

        with open("/tmp/spot_termination_notice", "w") as fp:
            fp.write(termination_time)

        command = [
            sys.executable,
            f"/metaflow/{flow_filename}",
            "spot-metadata",
            "record",
            "--run-id",
            run_id,
            "--step-name",
            step_name,
            "--task-id",
            task_id,
            "--termination-notice-time",
            termination_time,
            "--tag",
            "attempt_id:{}".format(retry_count),
        ]

        result = subprocess.run(command, capture_output=True, text=True)

        if result.returncode != 0:
            print(f"Failed to record spot termination metadata: {result.stderr}")



import importlib
from metaflow._vendor import click
from metaflow.extension_support.plugins import get_plugin


class LazyPluginCommandCollection(click.CommandCollection):
    # lazy_source should only point to things that are resolved as CLI plugins.
    def __init__(self, *args, lazy_sources=None, **kwargs):
        super().__init__(*args, **kwargs)
        # lazy_sources is a list of strings in the form
        # "{plugin_name}" -> "{module-name}.{command-object-name}"
        self.lazy_sources = lazy_sources or {}
        self._lazy_loaded = {}

    def invoke(self, ctx):
        # NOTE: This is copied from MultiCommand.invoke. The change is that we
        # behave like chain in the sense that we evaluate the subcommand *after*
        # invoking the base command but we don't chain the commands like self.chain
        # would otherwise indicate.
        # The goal of this is to make sure that the first command is properly executed
        # *first* prior to loading the other subcommands. It's more a lazy_subcommand_load
        # than a chain.
        # Look for CHANGE HERE in this code to see where the changes are made.
        # If click is updated, this may also need to be updated. This version is for
        # click 7.1.2.
        def _process_result(value):
            if self.result_callback is not None:
                value = ctx.invoke(self.result_callback, value, **ctx.params)
            return value

        if not ctx.protected_args:
            # If we are invoked without command the chain flag controls
            # how this happens.  If we are not in chain mode, the return
            # value here is the return value of the command.
            # If however we are in chain mode, the return value is the
            # return value of the result processor invoked with an empty
            # list (which means that no subcommand actually was executed).
            if self.invoke_without_command:
                # CHANGE HERE: We behave like self.chain = False here

                # if not self.chain:
                return click.Command.invoke(self, ctx)
                # with ctx:
                #    click.Command.invoke(self, ctx)
                #    return _process_result([])

            ctx.fail("Missing command.")

        # Fetch args back out
        args = ctx.protected_args + ctx.args
        ctx.args = []
        ctx.protected_args = []
        # CHANGE HERE: Add saved_args so we have access to it in the command to be
        # able to infer what we are calling next
        ctx.saved_args = args

        # If we're not in chain mode, we only allow the invocation of a
        # single command but we also inform the current context about the
        # name of the command to invoke.
        # CHANGE HERE: We change this block to do the invoke *before* the resolve_command
        # Make sure the context is entered so we do not clean up
        # resources until the result processor has worked.
        with ctx:
            ctx.invoked_subcommand = "*" if args else None
            click.Command.invoke(self, ctx)
            cmd_name, cmd, args = self.resolve_command(ctx, args)
            sub_ctx = cmd.make_context(cmd_name, args, parent=ctx)
            with sub_ctx:
                return _process_result(sub_ctx.command.invoke(sub_ctx))

        # CHANGE HERE: Removed all the part of chain mode.

    def list_commands(self, ctx):
        base = super().list_commands(ctx)
        for source_name, source in self.lazy_sources.items():
            subgroup = self._lazy_load(source_name, source)
            base.extend(subgroup.list_commands(ctx))
        return base

    def get_command(self, ctx, cmd_name):
        base_cmd = super().get_command(ctx, cmd_name)
        if base_cmd is not None:
            return base_cmd
        for source_name, source in self.lazy_sources.items():
            subgroup = self._lazy_load(source_name, source)
            cmd = subgroup.get_command(ctx, cmd_name)
            if cmd is not None:
                return cmd
        return None

    def _lazy_load(self, source_name, source_path):
        if source_name in self._lazy_loaded:
            return self._lazy_loaded[source_name]
        cmd_object = get_plugin("cli", source_path, source_name)
        if not isinstance(cmd_object, click.Group):
            raise ValueError(
                f"Lazy loading of {source_name} failed by returning "
                "a non-group object"
            )
        self._lazy_loaded[source_name] = cmd_object
        return cmd_object


class LazyGroup(click.Group):
    def __init__(self, *args, lazy_subcommands=None, **kwargs):
        super().__init__(*args, **kwargs)
        # lazy_subcommands is a list of strings in the form
        # "{command} -> "{module-name}.{command-object-name}"
        self.lazy_subcommands = lazy_subcommands or {}
        self._lazy_loaded = {}

    def list_commands(self, ctx):
        base = super().list_commands(ctx)
        lazy = sorted(self.lazy_subcommands.keys())
        return base + lazy

    def get_command(self, ctx, cmd_name):
        if cmd_name in self.lazy_subcommands:
            return self._lazy_load(cmd_name)
        return super().get_command(ctx, cmd_name)

    def _lazy_load(self, cmd_name):
        if cmd_name in self._lazy_loaded:
            return self._lazy_loaded[cmd_name]

        import_path = self.lazy_subcommands[cmd_name]
        modname, cmd = import_path.rsplit(".", 1)
        # do the import
        mod = importlib.import_module(modname)
        # get the Command object from that module
        cmd_object = getattr(mod, cmd)
        # check the result to make debugging easier. note that wrapped BaseCommand
        # can be functions
        if not isinstance(cmd_object, click.BaseCommand):
            raise ValueError(
                f"Lazy loading of {import_path} failed by returning "
                f"a non-command object {type(cmd_object)}"
            )
        self._lazy_loaded[cmd_name] = cmd_object
        return cmd_object



import sys
from metaflow.metaflow_config import (
    OTEL_ENDPOINT,
    ZIPKIN_ENDPOINT,
    CONSOLE_TRACE_ENABLED,
    SERVICE_AUTH_KEY,
    SERVICE_HEADERS,
)


def get_span_exporter():
    exporter_map = {
        OTEL_ENDPOINT: _create_otel_exporter,
        ZIPKIN_ENDPOINT: _create_zipkin_exporter,
        CONSOLE_TRACE_ENABLED: _create_console_exporter,
    }

    for config, create_exporter in exporter_map.items():
        if config:
            return create_exporter()

    print("WARNING: endpoints not set up for OpenTelemetry", file=sys.stderr)
    return None


def _create_otel_exporter():
    from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter

    if not any([SERVICE_AUTH_KEY, SERVICE_HEADERS]):
        print("WARNING: no auth settings for OpenTelemetry", file=sys.stderr)
        return None

    config = {
        "endpoint": OTEL_ENDPOINT,
        "timeout": 1,
    }

    if SERVICE_AUTH_KEY:
        config["headers"] = {"x-api-key": SERVICE_AUTH_KEY}
    elif SERVICE_HEADERS:
        config["headers"] = SERVICE_HEADERS

    return OTLPSpanExporter(**config)


def _create_zipkin_exporter():
    from opentelemetry.exporter.zipkin.proto.http import ZipkinExporter

    return ZipkinExporter(endpoint=ZIPKIN_ENDPOINT)


def _create_console_exporter():
    from opentelemetry.sdk.trace.export import ConsoleSpanExporter

    return ConsoleSpanExporter()



import copy
import errno
import fcntl
import functools
import io
import json
import os
import tarfile
import threading
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import wraps
from hashlib import sha256
from io import BufferedIOBase, BytesIO
from urllib.parse import unquote, urlparse

from metaflow.debug import debug
from metaflow.exception import MetaflowException
from metaflow.metaflow_config import get_pinned_conda_libs
from metaflow.metaflow_environment import MetaflowEnvironment
from metaflow.packaging_sys import ContentType

from . import MAGIC_FILE, _datastore_packageroot
from .utils import conda_platform


class CondaEnvironmentException(MetaflowException):
    headline = "Ran into an error while setting up environment"

    def __init__(self, msg):
        super(CondaEnvironmentException, self).__init__(msg)


class CondaEnvironment(MetaflowEnvironment):
    TYPE = "conda"
    _filecache = None
    _force_rebuild = False

    def __init__(self, flow):
        super().__init__(flow)
        self.flow = flow

    def set_local_root(self, local_root):
        # TODO: Make life simple by passing echo to the constructor and getting rid of
        # this method's invocation in the decorator
        self.local_root = local_root

    def decospecs(self):
        # Apply conda decorator to manage the task execution lifecycle.
        return ("conda",) + super().decospecs()

    def validate_environment(self, logger, datastore_type):
        self.datastore_type = datastore_type

        # Avoiding circular imports.
        from metaflow.plugins import DATASTORES

        self.datastore = [d for d in DATASTORES if d.TYPE == self.datastore_type][0]

        # Initialize necessary virtual environments for all Metaflow tasks.
        # Use Micromamba for solving conda packages and Pip for solving pypi packages.
        from .micromamba import Micromamba
        from .pip import Pip

        print_lock = threading.Lock()

        def make_thread_safe(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                with print_lock:
                    return func(*args, **kwargs)

            return wrapper

        self.logger = make_thread_safe(logger)

        # TODO: Wire up logging
        micromamba = Micromamba(self.logger, self._force_rebuild)
        self.solvers = {"conda": micromamba, "pypi": Pip(micromamba, self.logger)}

    def init_environment(self, echo, only_steps=None):
        # The implementation optimizes for latency to ensure as many operations can
        # be turned into cheap no-ops as feasible. Otherwise, we focus on maintaining
        # a balance between latency and maintainability of code without re-implementing
        # the internals of Micromamba and Pip.

        # TODO: Introduce verbose logging
        #       https://github.com/Netflix/metaflow/issues/1494

        def environments(type_):
            seen = set()
            for step in self.flow:
                if only_steps and step.name not in only_steps:
                    continue
                environment = self.get_environment(step)
                if type_ in environment and environment["id_"] not in seen:
                    seen.add(environment["id_"])
                    for platform in environment[type_]["platforms"]:
                        yield environment["id_"], {
                            **{
                                k: v
                                for k, v in environment[type_].items()
                                if k != "platforms"
                            },
                            **{"platform": platform},
                        }

        def solve(id_, environment, type_):
            # Cached solve - should be quick!
            platform = environment["platform"]
            return (
                id_,
                (
                    (
                        not self._force_rebuild
                        and self.read_from_environment_manifest([id_, platform, type_])
                    )
                    or self.write_to_environment_manifest(
                        [id_, platform, type_],
                        self.solvers[type_].solve(id_, **environment),
                    )
                ),
                environment["python"],
                platform,
            )

        def cache(storage, results, type_):
            debug.conda_exec(
                "Caching packages for %s environments %s"
                % (type_, [result[0] for result in results])
            )

            def _path(url, local_path):
                # Special handling for VCS packages
                if url.startswith("git+"):
                    base, _ = os.path.split(urlparse(url).path)
                    _, file = os.path.split(local_path)
                    prefix = url.split("@")[-1]
                    return urlparse(url).netloc + os.path.join(
                        unquote(base), prefix, file
                    )
                else:
                    return urlparse(url).netloc + urlparse(url).path

            local_packages = {
                url: {
                    # Path to package in datastore.
                    "path": _path(
                        url, local_path
                    ),  # urlparse(url).netloc + urlparse(url).path,
                    # Path to package on local disk.
                    "local_path": local_path,
                }
                for result in results
                for url, local_path in self.solvers[type_].metadata(*result).items()
            }
            dirty = set()
            # Prune list of packages to cache.

            _meta = copy.deepcopy(local_packages)
            for id_, packages, _, _ in results:
                for package in packages:
                    if package.get("path") and not self._force_rebuild:
                        # Cache only those packages that manifest is unaware of
                        local_packages.pop(package["url"], None)
                    else:
                        package["path"] = _meta[package["url"]]["path"]
                        dirty.add(id_)

            list_of_path_and_filehandle = [
                (
                    package["path"],
                    # Lazily fetch package from the interweb if needed.
                    # TODO: Depending on the len_hint, the package might be downloaded from
                    #       the interweb prematurely. save_bytes needs to be adjusted to handle
                    #       this scenario.
                    LazyOpen(
                        package["local_path"],
                        "rb",
                        url,
                    ),
                )
                for url, package in local_packages.items()
            ]
            debug.conda_exec(
                "Caching %s new packages to the datastore for %s environment %s"
                % (
                    len(list_of_path_and_filehandle),
                    type_,
                    [result[0] for result in results],
                )
            )
            storage.save_bytes(
                list_of_path_and_filehandle,
                len_hint=len(list_of_path_and_filehandle),
                overwrite=self._force_rebuild,
            )
            for id_, packages, _, platform in results:
                if id_ in dirty:
                    self.write_to_environment_manifest([id_, platform, type_], packages)

            debug.conda_exec("Finished caching packages.")

        storage = None
        if self.datastore_type not in ["local"]:
            # Initialize storage for caching if using a remote datastore
            storage = self.datastore(_datastore_packageroot(self.datastore, echo))

        self.logger("Bootstrapping virtual environment(s) ...")
        # Sequence of operations:
        #  1. Start all conda solves in parallel
        #  2. Download conda packages sequentially
        #  3. Create and cache conda environments in parallel
        #  4. Start PyPI solves in parallel after each conda environment is created
        #  5. Download PyPI packages sequentially
        #  6. Create and cache PyPI environments in parallel
        with ThreadPoolExecutor() as executor:
            # Start all conda solves in parallel
            debug.conda_exec("Solving packages for Conda environments..")
            conda_solve_futures = [
                executor.submit(lambda x: solve(*x, "conda"), env)
                for env in environments("conda")
            ]
            conda_create_futures = []

            pypi_envs = {env[0]: env for env in environments("pypi")}
            pypi_solve_futures = []
            pypi_create_futures = []

            cache_futures = []
            # Process conda results sequentially for downloads
            for future in as_completed(conda_solve_futures):
                result = future.result()
                # Sequential conda download
                debug.conda_exec(
                    "Downloading packages for Conda environment %s" % result[0]
                )
                self.solvers["conda"].download(*result)
                # Parallel conda create and cache
                conda_create_future = executor.submit(
                    self.solvers["conda"].create, *result
                )
                if storage:
                    cache_futures.append(
                        executor.submit(cache, storage, [result], "conda")
                    )

                # Queue PyPI solve to start after conda create
                if result[0] in pypi_envs:
                    debug.conda_exec(
                        "Solving packages for PyPI environment %s" % result[0]
                    )
                    # solve pypi envs uniquely
                    pypi_env = pypi_envs.pop(result[0])

                    def pypi_solve(env):
                        conda_create_future.result()  # Wait for conda create
                        return solve(*env, "pypi")

                    pypi_solve_futures.append(executor.submit(pypi_solve, pypi_env))
                else:
                    # add conda create future to the generic list
                    conda_create_futures.append(conda_create_future)

            # Process PyPI results sequentially for downloads
            for solve_future in as_completed(pypi_solve_futures):
                result = solve_future.result()
                # Sequential PyPI download
                debug.conda_exec(
                    "Downloading packages for PyPI environment %s" % result[0]
                )
                self.solvers["pypi"].download(*result)
                # Parallel PyPI create and cache
                pypi_create_futures.append(
                    executor.submit(self.solvers["pypi"].create, *result)
                )
                if storage:
                    cache_futures.append(
                        executor.submit(cache, storage, [result], "pypi")
                    )

            # Raise exceptions for conda create
            debug.conda_exec("Checking results for Conda create..")
            for future in as_completed(conda_create_futures):
                future.result()

            # Raise exceptions for pypi create
            debug.conda_exec("Checking results for PyPI create..")
            for future in as_completed(pypi_create_futures):
                future.result()

            # Raise exceptions for caching
            debug.conda_exec("Checking results for caching..")
            for future in as_completed(cache_futures):
                # check for result in order to raise any exceptions.
                future.result()

        self.logger("Virtual environment(s) bootstrapped!")

    def executable(self, step_name, default=None):
        step = next((step for step in self.flow if step.name == step_name), None)
        if step is None:
            # requesting internal steps e.g. _parameters
            return super().executable(step_name, default)
        id_ = self.get_environment(step).get("id_")
        if id_:
            # bootstrap.py is responsible for ensuring the validity of this executable.
            # -s is important! Can otherwise leak packages to other environments.
            return os.path.join("$MF_ARCH", id_, "bin/python -s")
        else:
            # for @conda/@pypi(disabled=True).
            return super().executable(step_name, default)

    def interpreter(self, step_name):
        step = next(step for step in self.flow if step.name == step_name)
        id_ = self.get_environment(step)["id_"]
        # User workloads are executed through the conda environment's interpreter.
        return self.solvers["conda"].interpreter(id_)

    def is_disabled(self, step):
        for decorator in step.decorators:
            # @conda decorator is guaranteed to exist thanks to self.decospecs
            if decorator.name in ["conda", "pypi"]:
                # handle @conda/@pypi(disabled=True)
                disabled = decorator.attributes["disabled"]
                return str(disabled).lower() == "true"
        return False

    @functools.lru_cache(maxsize=None)
    def get_environment(self, step):
        environment = {}
        for decorator in step.decorators:
            # @conda decorator is guaranteed to exist thanks to self.decospecs
            if decorator.name in ["conda", "pypi"]:
                # handle @conda/@pypi(disabled=True)
                disabled = decorator.attributes["disabled"]
                if not disabled or str(disabled).lower() == "false":
                    environment[decorator.name] = {
                        k: copy.deepcopy(decorator.attributes[k])
                        for k in decorator.attributes
                        if k not in ("disabled", "libraries")
                    }
                else:
                    return {}
        # Resolve conda environment for @pypi's Python, falling back on @conda's
        # Python
        env_python = (
            environment.get("pypi", environment["conda"]).get("python")
            or environment["conda"]["python"]
        )
        # TODO: Support dependencies for `--metadata`.
        # TODO: Introduce support for `--telemetry` as a follow up.
        # Certain packages are required for metaflow runtime to function correctly.
        # Ensure these packages are available both in Conda channels and PyPI
        # repostories.
        pinned_packages = get_pinned_conda_libs(env_python, self.datastore_type)

        # PyPI dependencies are prioritized over Conda dependencies.
        environment.get("pypi", environment["conda"])["packages"] = {
            **pinned_packages,
            **environment.get("pypi", environment["conda"])["packages"],
        }
        # Disallow specifying both @conda and @pypi together for now. Mixing Conda
        # and PyPI packages comes with a lot of operational pain that we can handle
        # as follow-up work in the future.
        if all(
            map(lambda key: environment.get(key, {}).get("packages"), ["pypi", "conda"])
        ):
            msg = "Mixing and matching PyPI packages and Conda packages within a\n"
            msg += "step is not yet supported. Use one of @pypi or @conda only."
            raise CondaEnvironmentException(msg)

        # To support cross-platform environments, these invariants are maintained
        # 1. Conda packages are resolved for target platforms
        # 2. Conda packages are resolved for local platform only for PyPI packages
        # 3. Conda environments are created only for local platform
        # 4. PyPI packages are resolved for target platform within Conda environments
        #    created for local platform
        # 5. All resolved packages (Conda or PyPI) are cached
        # 6. PyPI packages are only installed for local platform

        target_platform = conda_platform()
        for decorator in step.decorators:
            # NOTE: Keep the list of supported decorator names for backward compatibility purposes.
            # Older versions did not implement the 'support_conda_environment' attribute.
            if getattr(
                decorator, "supports_conda_environment", False
            ) or decorator.name in [
                "batch",
                "kubernetes",
                "nvidia",
                "snowpark",
                "slurm",
                "nvct",
            ]:
                target_platform = getattr(decorator, "target_platform", "linux-64")
                break

        environment["conda"]["platforms"] = [target_platform]
        if "pypi" in environment:
            # For PyPI packages, resolve conda environment for local platform in
            # addition to target platform
            environment["conda"]["platforms"] = list(
                {target_platform, conda_platform()}
            )
            environment["pypi"]["platforms"] = [target_platform]
            # Match PyPI and Conda python versions with the resolved environment Python.
            environment["pypi"]["python"] = environment["conda"]["python"] = env_python

            # When using `Application Default Credentials` for private GCP
            # PyPI registries, the usage of environment variable `GOOGLE_APPLICATION_CREDENTIALS`
            # demands that `keyrings.google-artifactregistry-auth` has to be installed
            # and available in the underlying python environment.
            if os.getenv("GOOGLE_APPLICATION_CREDENTIALS"):
                environment["conda"]["packages"][
                    "keyrings.google-artifactregistry-auth"
                ] = ">=1.1.1"

        # Z combinator for a recursive lambda
        deep_sort = (lambda f: f(f))(
            lambda f: lambda obj: (
                {k: f(f)(v) for k, v in sorted(obj.items())}
                if isinstance(obj, dict)
                else sorted([f(f)(e) for e in obj]) if isinstance(obj, list) else obj
            )
        )

        return {
            **environment,
            # Create a stable unique id for the environment.
            # Add packageroot to the id so that packageroot modifications can
            # invalidate existing environments.
            "id_": sha256(
                json.dumps(
                    deep_sort(
                        {
                            **environment,
                            **{
                                "package_root": _datastore_packageroot(
                                    self.datastore, self.logger
                                )
                            },
                        }
                    )
                ).encode()
            ).hexdigest()[:15],
        }

    def pylint_config(self):
        config = super().pylint_config()
        # Disable (import-error) in pylint
        config.append("--disable=F0401")
        return config

    @classmethod
    def get_client_info(cls, flow_name, metadata):
        if cls._filecache is None:
            from metaflow.client.filecache import FileCache

            cls._filecache = FileCache()

        info = metadata.get("code-package")
        prefix = metadata.get("conda_env_prefix")
        if info is None or prefix is None:
            return {}
        info = json.loads(info)
        _, blobdata = cls._filecache.get_data(
            info["ds_type"], flow_name, info["location"], info["sha"]
        )
        with tarfile.open(fileobj=BytesIO(blobdata), mode="r:gz") as tar:
            manifest = tar.extractfile(MAGIC_FILE)
            info = json.loads(manifest.read().decode("utf-8"))
            return info[prefix.split("/")[2]][prefix.split("/")[1]]

    def add_to_package(self):
        # Add manifest file to job package at the top level.
        files = []
        manifest = self.get_environment_manifest_path()
        if os.path.exists(manifest):
            files.append(
                (manifest, os.path.basename(manifest), ContentType.OTHER_CONTENT)
            )
        return files

    def bootstrap_commands(self, step_name, datastore_type):
        # Bootstrap conda and execution environment for step
        step = next(step for step in self.flow if step.name == step_name)
        id_ = self.get_environment(step).get("id_")
        if id_:
            return [
                "echo 'Bootstrapping virtual environment...'",
                "flush_mflogs",
                # We have to prevent the tracing module from loading,
                # as the bootstrapping process uses the internal S3 client which would fail to import tracing
                # due to the required dependencies being bundled into the conda environment,
                # which is yet to be initialized at this point.
                'DISABLE_TRACING=True python -m metaflow.plugins.pypi.bootstrap "%s" %s "%s"'
                % (self.flow.name, id_, self.datastore_type),
                "echo 'Environment bootstrapped.'",
                "flush_mflogs",
                # To avoid having to install micromamba in the PATH in micromamba.py, we add it to the PATH here.
                "export PATH=$PATH:$(pwd)/micromamba/bin",
                "export MF_ARCH=$(case $(uname)/$(uname -m) in Darwin/arm64)echo osx-arm64;;Darwin/*)echo osx-64;;Linux/aarch64)echo linux-aarch64;;*)echo linux-64;;esac)",
            ]
        else:
            # for @conda/@pypi(disabled=True).
            return super().bootstrap_commands(step_name, datastore_type)

    # TODO: Make this an instance variable once local_root is part of the object
    #       constructor.
    def get_environment_manifest_path(self):
        return os.path.join(self.local_root, self.flow.name, MAGIC_FILE)

    def read_from_environment_manifest(self, keys):
        path = self.get_environment_manifest_path()
        if os.path.exists(path) and os.path.getsize(path) > 0:
            with open(path) as f:
                data = json.load(f)
                for key in keys:
                    try:
                        data = data[key]
                    except KeyError:
                        return None
                return data

    def write_to_environment_manifest(self, keys, value):
        path = self.get_environment_manifest_path()
        try:
            os.makedirs(os.path.dirname(path))
        except OSError as x:
            if x.errno != errno.EEXIST:
                raise
        with os.fdopen(os.open(path, os.O_RDWR | os.O_CREAT), "r+") as f:
            try:
                fcntl.flock(f, fcntl.LOCK_EX)
                d = {}
                if os.path.getsize(path) > 0:
                    f.seek(0)
                    d = json.load(f)
                data = d
                for key in keys[:-1]:
                    data = data.setdefault(key, {})
                data[keys[-1]] = value
                f.seek(0)
                json.dump(d, f)
                f.truncate()
                return value
            except IOError as e:
                if e.errno != errno.EAGAIN:
                    raise
            finally:
                fcntl.flock(f, fcntl.LOCK_UN)


class LazyOpen(BufferedIOBase):
    def __init__(self, filename, mode="rb", url=None):
        super().__init__()
        self.filename = filename
        self.mode = mode
        self.url = url
        self._file = None
        self._buffer = None
        self._position = 0
        self.requests = None

    def _ensure_file(self):
        if not self._file:
            if self.filename and os.path.exists(self.filename):
                self._file = open(self.filename, self.mode)
            elif self.url:
                if self.url.startswith("git+"):
                    raise ValueError(
                        "LazyOpen doesn't support VCS url %s yet!" % self.url
                    )
                self._buffer = self._download_to_buffer()
                self._file = io.BytesIO(self._buffer)
            else:
                raise ValueError("Both filename and url are missing")

    def _download_to_buffer(self):
        if self.requests is None:
            # TODO: Remove dependency on requests
            import requests

            self.requests = requests
        # TODO: Stream it in chunks?
        response = self.requests.get(self.url, stream=True)
        response.raise_for_status()
        return response.content

    def readable(self):
        return "r" in self.mode

    def seekable(self):
        return True

    def read(self, size=-1):
        self._ensure_file()
        return self._file.read(size)

    def seek(self, offset, whence=io.SEEK_SET):
        self._ensure_file()
        return self._file.seek(offset, whence)

    def tell(self):
        self._ensure_file()
        return self._file.tell()

    def close(self):
        if self._file:
            self._file.close()



from typing import Dict, Optional, List, Tuple


class _TrieNode:
    def __init__(
        self, parent: Optional["_TrieNode"] = None, component: Optional[str] = None
    ):
        self.parent = parent
        self.component = component
        self.children = {}  # type: Dict[str, "_TrieNode"]
        self.total_children = 0
        self.value = None
        self.end_value = None

    def traverse(self, value: type) -> Optional["_TrieNode"]:
        if self.total_children == 0:
            self.end_value = value
        else:
            self.end_value = None
        self.total_children += 1

    def remove_child(self, child_name: str) -> bool:
        if child_name in self.children:
            del self.children[child_name]
            self.total_children -= 1
            return True
        return False


class ClassPath_Trie:
    def __init__(self):
        self.root = _TrieNode(None, None)
        self.inited = False
        self._value_to_node = {}  # type: Dict[type, _TrieNode]

    def init(self, initial_nodes: Optional[List[Tuple[str, type]]] = None):
        # We need to do this so we can delay import of STEP_DECORATORS
        self.inited = True
        for classpath_name, value in initial_nodes or []:
            self.insert(classpath_name, value)

    def insert(self, classpath_name: str, value: type):
        node = self.root
        components = reversed(classpath_name.split("."))
        for c in components:
            node = node.children.setdefault(c, _TrieNode(node, c))
            node.traverse(value)
        node.total_children -= (
            1  # We do not count the last node as having itself as a child
        )
        node.value = value
        self._value_to_node[value] = node

    def search(self, classpath_name: str) -> Optional[type]:
        node = self.root
        components = reversed(classpath_name.split("."))
        for c in components:
            if c not in node.children:
                return None
            node = node.children[c]
        return node.value

    def remove(self, classpath_name: str):
        components = list(reversed(classpath_name.split(".")))

        def _remove(node: _TrieNode, components, depth):
            if depth == len(components):
                if node.value is not None:
                    del self._value_to_node[node.value]
                    node.value = None
                    return len(node.children) == 0
                return False
            c = components[depth]
            if c not in node.children:
                return False
            did_delete_child = _remove(node.children[c], components, depth + 1)
            if did_delete_child:
                node.remove_child(c)
                if node.total_children == 1:
                    # If we have one total child left, we have at least one
                    # child and that one has an end_value
                    for child in node.children.values():
                        assert (
                            child.end_value
                        ), "Node with one child must have an end_value"
                        node.end_value = child.end_value
                return node.total_children == 0
            return False

        _remove(self.root, components, 0)

    def unique_prefix_value(self, classpath_name: str) -> Optional[type]:
        node = self.root
        components = reversed(classpath_name.split("."))
        for c in components:
            if c not in node.children:
                return None
            node = node.children[c]
        # If we reach here, it means the classpath_name is a prefix.
        # We check if it has only one path forward (end_value will be non None)
        # If value is not None, we also consider this to be a unique "prefix"
        # This happens since this trie is also filled with metaflow default decorators
        return node.end_value or node.value

    def unique_prefix_for_type(self, value: type) -> Optional[str]:
        node = self._value_to_node.get(value, None)
        if node is None:
            return None
        components = []
        while node:
            if node.end_value == value:
                components = []
            if node.component is not None:
                components.append(node.component)
            node = node.parent
        return ".".join(components)

    def get_unique_prefixes(self) -> Dict[str, type]:
        """
        Get all unique prefixes in the trie.

        Returns
        -------
        List[str]
            A list of unique prefixes.
        """
        to_return = {}

        def _collect(node, current_prefix):
            if node.end_value is not None:
                to_return[current_prefix] = node.end_value
                # We stop there and don't look further since we found the unique prefix
                return
            if node.value is not None:
                to_return[current_prefix] = node.value
                # We continue to look for more unique prefixes
            for child_name, child_node in node.children.items():
                _collect(
                    child_node,
                    f"{current_prefix}.{child_name}" if current_prefix else child_name,
                )

        _collect(self.root, "")
        return {".".join(reversed(k.split("."))): v for k, v in to_return.items()}



from metaflow.decorators import StepDecorator


class EnvironmentDecorator(StepDecorator):
    """
    Specifies environment variables to be set prior to the execution of a step.

    Parameters
    ----------
    vars : Dict[str, str], default {}
        Dictionary of environment variables to set.
    """

    name = "environment"
    defaults = {"vars": {}}

    def runtime_step_cli(
        self, cli_args, retry_count, max_user_code_retries, ubf_context
    ):
        cli_args.env.update(
            {key: str(value) for key, value in self.attributes["vars"].items()}
        )



from __future__ import annotations

import ast
import sys
import types
from collections.abc import Callable, Iterable
from importlib.abc import MetaPathFinder
from importlib.machinery import ModuleSpec, SourceFileLoader
from importlib.util import cache_from_source, decode_source
from inspect import isclass
from os import PathLike
from types import CodeType, ModuleType, TracebackType
from typing import Sequence, TypeVar
from unittest.mock import patch

from ._config import global_config
from ._transformer import TypeguardTransformer

if sys.version_info >= (3, 12):
    from collections.abc import Buffer
else:
    from metaflow._vendor.v3_7.typing_extensions import Buffer

if sys.version_info >= (3, 11):
    from typing import ParamSpec
else:
    from metaflow._vendor.v3_7.typing_extensions import ParamSpec

if sys.version_info >= (3, 10):
    from importlib.metadata import PackageNotFoundError, version
else:
    from metaflow._vendor.v3_7.importlib_metadata import PackageNotFoundError, version

try:
    OPTIMIZATION = "typeguard" + "".join(version("typeguard").split(".")[:3])
except PackageNotFoundError:
    OPTIMIZATION = "typeguard"

P = ParamSpec("P")
T = TypeVar("T")


# The name of this function is magical
def _call_with_frames_removed(
    f: Callable[P, T], *args: P.args, **kwargs: P.kwargs
) -> T:
    return f(*args, **kwargs)


def optimized_cache_from_source(path: str, debug_override: bool | None = None) -> str:
    return cache_from_source(path, debug_override, optimization=OPTIMIZATION)


class TypeguardLoader(SourceFileLoader):
    @staticmethod
    def source_to_code(
        data: Buffer | str | ast.Module | ast.Expression | ast.Interactive,
        path: Buffer | str | PathLike[str] = "<string>",
    ) -> CodeType:
        if isinstance(data, (ast.Module, ast.Expression, ast.Interactive)):
            tree = data
        else:
            if isinstance(data, str):
                source = data
            else:
                source = decode_source(data)

            tree = _call_with_frames_removed(
                ast.parse,
                source,
                path,
                "exec",
            )

        tree = TypeguardTransformer().visit(tree)
        ast.fix_missing_locations(tree)

        if global_config.debug_instrumentation and sys.version_info >= (3, 9):
            print(
                f"Source code of {path!r} after instrumentation:\n"
                "----------------------------------------------",
                file=sys.stderr,
            )
            print(ast.unparse(tree), file=sys.stderr)
            print("----------------------------------------------", file=sys.stderr)

        return _call_with_frames_removed(
            compile, tree, path, "exec", 0, dont_inherit=True
        )

    def exec_module(self, module: ModuleType) -> None:
        # Use a custom optimization marker – the import lock should make this monkey
        # patch safe
        with patch(
            "importlib._bootstrap_external.cache_from_source",
            optimized_cache_from_source,
        ):
            super().exec_module(module)


class TypeguardFinder(MetaPathFinder):
    """
    Wraps another path finder and instruments the module with
    :func:`@typechecked <typeguard.typechecked>` if :meth:`should_instrument` returns
    ``True``.

    Should not be used directly, but rather via :func:`~.install_import_hook`.

    .. versionadded:: 2.6
    """

    def __init__(self, packages: list[str] | None, original_pathfinder: MetaPathFinder):
        self.packages = packages
        self._original_pathfinder = original_pathfinder

    def find_spec(
        self,
        fullname: str,
        path: Sequence[str] | None,
        target: types.ModuleType | None = None,
    ) -> ModuleSpec | None:
        if self.should_instrument(fullname):
            spec = self._original_pathfinder.find_spec(fullname, path, target)
            if spec is not None and isinstance(spec.loader, SourceFileLoader):
                spec.loader = TypeguardLoader(spec.loader.name, spec.loader.path)
                return spec

        return None

    def should_instrument(self, module_name: str) -> bool:
        """
        Determine whether the module with the given name should be instrumented.

        :param module_name: full name of the module that is about to be imported (e.g.
            ``xyz.abc``)

        """
        if self.packages is None:
            return True

        for package in self.packages:
            if module_name == package or module_name.startswith(package + "."):
                return True

        return False


class ImportHookManager:
    """
    A handle that can be used to uninstall the Typeguard import hook.
    """

    def __init__(self, hook: MetaPathFinder):
        self.hook = hook

    def __enter__(self) -> None:
        pass

    def __exit__(
        self,
        exc_type: type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        self.uninstall()

    def uninstall(self) -> None:
        """Uninstall the import hook."""
        try:
            sys.meta_path.remove(self.hook)
        except ValueError:
            pass  # already removed


def install_import_hook(
    packages: Iterable[str] | None = None,
    *,
    cls: type[TypeguardFinder] = TypeguardFinder,
) -> ImportHookManager:
    """
    Install an import hook that instruments functions for automatic type checking.

    This only affects modules loaded **after** this hook has been installed.

    :param packages: an iterable of package names to instrument, or ``None`` to
        instrument all packages
    :param cls: a custom meta path finder class
    :return: a context manager that uninstalls the hook on exit (or when you call
        ``.uninstall()``)

    .. versionadded:: 2.6

    """
    if packages is None:
        target_packages: list[str] | None = None
    elif isinstance(packages, str):
        target_packages = [packages]
    else:
        target_packages = list(packages)

    for finder in sys.meta_path:
        if (
            isclass(finder)
            and finder.__name__ == "PathFinder"
            and hasattr(finder, "find_spec")
        ):
            break
    else:
        raise RuntimeError("Cannot find a PathFinder in sys.meta_path")

    hook = cls(target_packages, finder)
    sys.meta_path.insert(0, hook)
    return ImportHookManager(hook)



from metaflow.exception import MetaflowException
from metaflow.metaflow_config import AZURE_STORAGE_BLOB_SERVICE_ENDPOINT
from metaflow.plugins.azure.azure_utils import (
    check_azure_deps,
)
from metaflow.plugins.azure.azure_credential import (
    create_cacheable_azure_credential,
)

import os
import threading


class _ClientCache(object):
    """
    azure.storage.blob.BlobServiceClient objects internally cache HTTPS connections.
    In order to reuse HTTPS connections, we need to reuse BlobServiceClient objects.

    The effect we are going for is for each process or thread in the executor to
    EACH REUSE ITS OWN long-lived BlobServiceClient object.
    """

    _cache = dict()

    def __init__(self):
        raise RuntimeError("_ClientCache may not be instantiated!")

    @staticmethod
    def get(
        blob_service_endpoint,
        credential=None,
        max_single_put_size=None,
        max_single_get_size=None,
        max_chunk_get_size=None,
        connection_data_block_size=None,
    ):
        """
        Each spawned process will get its own fresh cache.

        With each process's cache, each thread gets its own distinct service object.

        The cache key includes all BlobServiceClient creation params PLUS process ID and thread ID.

        This means that no more than one thread of control will ever access a cache key. This, and
        the fact that dict() operations are thread-safe means that no additional synchronization
        required here.
        """
        cache_key = (
            os.getpid(),
            threading.get_ident(),
            credential,
            blob_service_endpoint,
            max_single_put_size,
            max_single_get_size,
            max_chunk_get_size,
            connection_data_block_size,
        )
        service_just_created = None
        if cache_key not in _ClientCache._cache:
            service_just_created = _create_blob_service_client(
                blob_service_endpoint,
                credential=credential,
                max_single_put_size=max_single_put_size,
                max_single_get_size=max_single_get_size,
                max_chunk_get_size=max_chunk_get_size,
                connection_data_block_size=connection_data_block_size,
                # BlobServiceClient accepts many more kwargs. We can add passthroughs as/when needed.
            )
            _ClientCache._cache[cache_key] = service_just_created
        service_to_return = _ClientCache._cache[cache_key]
        if (
            service_just_created is not None
            and service_just_created != service_to_return
        ):
            # This condition may not be fatal, but highly unexpected.
            # Each thread of execution gets its own entry, so there *should* be no concurrent insertions
            # on the same key.
            #
            # The Metaflow team REALLY wants to know if this ever happens, but we stop short of raising a
            # fatal exception to not kill user jobs.
            print(
                "METAFLOW WARNING: Azure _ClientCache had the same cache key updated more than once"
            )
        return service_to_return


# AZURE SDK tunables (AZURE_CLIENT_*)
# ===================================
# We try pick some sensible defaults here.
#
# Block size on underlying down TLS transport - Azure SDK defaults to 4096.
# This larger block size dramatically improves aggregate download throughput.
# 256KB (within order of 2x) seemed optimal for specific benchmark setup.
# In the future, exposing this as a tunable to users is a possibility if necessary.
AZURE_CLIENT_CONNECTION_DATA_BLOCK_SIZE = 262144

# When to use more than a single thread / connection for a single GET or PUT.
AZURE_CLIENT_MAX_SINGLE_GET_SIZE_MB = 32
AZURE_CLIENT_MAX_SINGLE_PUT_SIZE_MB = 64

# Maximum chunk size when splitting a single blob GET into chunks for concurrent processing
AZURE_CLIENT_MAX_CHUNK_GET_SIZE_MB = 16

BYTES_IN_MB = 1024 * 1024


def get_azure_blob_service_client(
    credential=None,
    credential_is_cacheable=False,
    max_single_get_size=AZURE_CLIENT_MAX_SINGLE_GET_SIZE_MB * BYTES_IN_MB,
    max_single_put_size=AZURE_CLIENT_MAX_SINGLE_PUT_SIZE_MB * BYTES_IN_MB,
    max_chunk_get_size=AZURE_CLIENT_MAX_CHUNK_GET_SIZE_MB * BYTES_IN_MB,
    connection_data_block_size=AZURE_CLIENT_CONNECTION_DATA_BLOCK_SIZE,
):
    """Returns a azure.storage.blob.BlobServiceClient.

    The value adds are:
    - connection caching (see _ClientCache)
    - auto storage account URL detection
    - auto credential handling (pull SAS token from environment, OR DefaultAzureCredential)
    - sensible default values for Azure SDK tunables
    """
    if not AZURE_STORAGE_BLOB_SERVICE_ENDPOINT:
        raise MetaflowException(
            msg="Must configure METAFLOW_AZURE_STORAGE_BLOB_SERVICE_ENDPOINT"
        )
    blob_service_endpoint = AZURE_STORAGE_BLOB_SERVICE_ENDPOINT

    if not credential:
        credential = create_cacheable_azure_credential()
        credential_is_cacheable = True

    if not credential_is_cacheable:
        return _create_blob_service_client(
            blob_service_endpoint,
            credential=credential,
            max_single_put_size=max_single_put_size,
            max_single_get_size=max_single_get_size,
            max_chunk_get_size=max_chunk_get_size,
            connection_data_block_size=connection_data_block_size,
            # BlobServiceClient accepts many more kwargs. We can add passthroughs as/when needed.
        )

    return _ClientCache.get(
        blob_service_endpoint,
        credential=credential,
        max_single_put_size=max_single_put_size,
        max_single_get_size=max_single_get_size,
        max_chunk_get_size=max_chunk_get_size,
        connection_data_block_size=connection_data_block_size,
    )


@check_azure_deps
def _create_blob_service_client(
    blob_service_endpoint,
    credential=None,
    max_single_put_size=None,
    max_single_get_size=None,
    max_chunk_get_size=None,
    connection_data_block_size=None,
):
    from azure.storage.blob import BlobServiceClient

    return BlobServiceClient(
        blob_service_endpoint,
        credential=credential,
        max_single_put_size=max_single_put_size,
        max_single_get_size=max_single_get_size,
        max_chunk_get_size=max_chunk_get_size,
        connection_data_block_size=connection_data_block_size,
        # BlobServiceClient accepts many more kwargs. We can add passthroughs as/when needed.
    )




__all__ = ['BaseRepresenter', 'SafeRepresenter', 'Representer',
    'RepresenterError']

from .error import *
from .nodes import *

import datetime, copyreg, types, base64, collections

class RepresenterError(YAMLError):
    pass

class BaseRepresenter:

    yaml_representers = {}
    yaml_multi_representers = {}

    def __init__(self, default_style=None, default_flow_style=False, sort_keys=True):
        self.default_style = default_style
        self.sort_keys = sort_keys
        self.default_flow_style = default_flow_style
        self.represented_objects = {}
        self.object_keeper = []
        self.alias_key = None

    def represent(self, data):
        node = self.represent_data(data)
        self.serialize(node)
        self.represented_objects = {}
        self.object_keeper = []
        self.alias_key = None

    def represent_data(self, data):
        if self.ignore_aliases(data):
            self.alias_key = None
        else:
            self.alias_key = id(data)
        if self.alias_key is not None:
            if self.alias_key in self.represented_objects:
                node = self.represented_objects[self.alias_key]
                #if node is None:
                #    raise RepresenterError("recursive objects are not allowed: %r" % data)
                return node
            #self.represented_objects[alias_key] = None
            self.object_keeper.append(data)
        data_types = type(data).__mro__
        if data_types[0] in self.yaml_representers:
            node = self.yaml_representers[data_types[0]](self, data)
        else:
            for data_type in data_types:
                if data_type in self.yaml_multi_representers:
                    node = self.yaml_multi_representers[data_type](self, data)
                    break
            else:
                if None in self.yaml_multi_representers:
                    node = self.yaml_multi_representers[None](self, data)
                elif None in self.yaml_representers:
                    node = self.yaml_representers[None](self, data)
                else:
                    node = ScalarNode(None, str(data))
        #if alias_key is not None:
        #    self.represented_objects[alias_key] = node
        return node

    @classmethod
    def add_representer(cls, data_type, representer):
        if not 'yaml_representers' in cls.__dict__:
            cls.yaml_representers = cls.yaml_representers.copy()
        cls.yaml_representers[data_type] = representer

    @classmethod
    def add_multi_representer(cls, data_type, representer):
        if not 'yaml_multi_representers' in cls.__dict__:
            cls.yaml_multi_representers = cls.yaml_multi_representers.copy()
        cls.yaml_multi_representers[data_type] = representer

    def represent_scalar(self, tag, value, style=None):
        if style is None:
            style = self.default_style
        node = ScalarNode(tag, value, style=style)
        if self.alias_key is not None:
            self.represented_objects[self.alias_key] = node
        return node

    def represent_sequence(self, tag, sequence, flow_style=None):
        value = []
        node = SequenceNode(tag, value, flow_style=flow_style)
        if self.alias_key is not None:
            self.represented_objects[self.alias_key] = node
        best_style = True
        for item in sequence:
            node_item = self.represent_data(item)
            if not (isinstance(node_item, ScalarNode) and not node_item.style):
                best_style = False
            value.append(node_item)
        if flow_style is None:
            if self.default_flow_style is not None:
                node.flow_style = self.default_flow_style
            else:
                node.flow_style = best_style
        return node

    def represent_mapping(self, tag, mapping, flow_style=None):
        value = []
        node = MappingNode(tag, value, flow_style=flow_style)
        if self.alias_key is not None:
            self.represented_objects[self.alias_key] = node
        best_style = True
        if hasattr(mapping, 'items'):
            mapping = list(mapping.items())
            if self.sort_keys:
                try:
                    mapping = sorted(mapping)
                except TypeError:
                    pass
        for item_key, item_value in mapping:
            node_key = self.represent_data(item_key)
            node_value = self.represent_data(item_value)
            if not (isinstance(node_key, ScalarNode) and not node_key.style):
                best_style = False
            if not (isinstance(node_value, ScalarNode) and not node_value.style):
                best_style = False
            value.append((node_key, node_value))
        if flow_style is None:
            if self.default_flow_style is not None:
                node.flow_style = self.default_flow_style
            else:
                node.flow_style = best_style
        return node

    def ignore_aliases(self, data):
        return False

class SafeRepresenter(BaseRepresenter):

    def ignore_aliases(self, data):
        if data is None:
            return True
        if isinstance(data, tuple) and data == ():
            return True
        if isinstance(data, (str, bytes, bool, int, float)):
            return True

    def represent_none(self, data):
        return self.represent_scalar('tag:yaml.org,2002:null', 'null')

    def represent_str(self, data):
        return self.represent_scalar('tag:yaml.org,2002:str', data)

    def represent_binary(self, data):
        if hasattr(base64, 'encodebytes'):
            data = base64.encodebytes(data).decode('ascii')
        else:
            data = base64.encodestring(data).decode('ascii')
        return self.represent_scalar('tag:yaml.org,2002:binary', data, style='|')

    def represent_bool(self, data):
        if data:
            value = 'true'
        else:
            value = 'false'
        return self.represent_scalar('tag:yaml.org,2002:bool', value)

    def represent_int(self, data):
        return self.represent_scalar('tag:yaml.org,2002:int', str(data))

    inf_value = 1e300
    while repr(inf_value) != repr(inf_value*inf_value):
        inf_value *= inf_value

    def represent_float(self, data):
        if data != data or (data == 0.0 and data == 1.0):
            value = '.nan'
        elif data == self.inf_value:
            value = '.inf'
        elif data == -self.inf_value:
            value = '-.inf'
        else:
            value = repr(data).lower()
            # Note that in some cases `repr(data)` represents a float number
            # without the decimal parts.  For instance:
            #   >>> repr(1e17)
            #   '1e17'
            # Unfortunately, this is not a valid float representation according
            # to the definition of the `!!float` tag.  We fix this by adding
            # '.0' before the 'e' symbol.
            if '.' not in value and 'e' in value:
                value = value.replace('e', '.0e', 1)
        return self.represent_scalar('tag:yaml.org,2002:float', value)

    def represent_list(self, data):
        #pairs = (len(data) > 0 and isinstance(data, list))
        #if pairs:
        #    for item in data:
        #        if not isinstance(item, tuple) or len(item) != 2:
        #            pairs = False
        #            break
        #if not pairs:
            return self.represent_sequence('tag:yaml.org,2002:seq', data)
        #value = []
        #for item_key, item_value in data:
        #    value.append(self.represent_mapping(u'tag:yaml.org,2002:map',
        #        [(item_key, item_value)]))
        #return SequenceNode(u'tag:yaml.org,2002:pairs', value)

    def represent_dict(self, data):
        return self.represent_mapping('tag:yaml.org,2002:map', data)

    def represent_set(self, data):
        value = {}
        for key in data:
            value[key] = None
        return self.represent_mapping('tag:yaml.org,2002:set', value)

    def represent_date(self, data):
        value = data.isoformat()
        return self.represent_scalar('tag:yaml.org,2002:timestamp', value)

    def represent_datetime(self, data):
        value = data.isoformat(' ')
        return self.represent_scalar('tag:yaml.org,2002:timestamp', value)

    def represent_yaml_object(self, tag, data, cls, flow_style=None):
        if hasattr(data, '__getstate__'):
            state = data.__getstate__()
        else:
            state = data.__dict__.copy()
        return self.represent_mapping(tag, state, flow_style=flow_style)

    def represent_undefined(self, data):
        raise RepresenterError("cannot represent an object", data)

SafeRepresenter.add_representer(type(None),
        SafeRepresenter.represent_none)

SafeRepresenter.add_representer(str,
        SafeRepresenter.represent_str)

SafeRepresenter.add_representer(bytes,
        SafeRepresenter.represent_binary)

SafeRepresenter.add_representer(bool,
        SafeRepresenter.represent_bool)

SafeRepresenter.add_representer(int,
        SafeRepresenter.represent_int)

SafeRepresenter.add_representer(float,
        SafeRepresenter.represent_float)

SafeRepresenter.add_representer(list,
        SafeRepresenter.represent_list)

SafeRepresenter.add_representer(tuple,
        SafeRepresenter.represent_list)

SafeRepresenter.add_representer(dict,
        SafeRepresenter.represent_dict)

SafeRepresenter.add_representer(set,
        SafeRepresenter.represent_set)

SafeRepresenter.add_representer(datetime.date,
        SafeRepresenter.represent_date)

SafeRepresenter.add_representer(datetime.datetime,
        SafeRepresenter.represent_datetime)

SafeRepresenter.add_representer(None,
        SafeRepresenter.represent_undefined)

class Representer(SafeRepresenter):

    def represent_complex(self, data):
        if data.imag == 0.0:
            data = '%r' % data.real
        elif data.real == 0.0:
            data = '%rj' % data.imag
        elif data.imag > 0:
            data = '%r+%rj' % (data.real, data.imag)
        else:
            data = '%r%rj' % (data.real, data.imag)
        return self.represent_scalar('tag:yaml.org,2002:python/complex', data)

    def represent_tuple(self, data):
        return self.represent_sequence('tag:yaml.org,2002:python/tuple', data)

    def represent_name(self, data):
        name = '%s.%s' % (data.__module__, data.__name__)
        return self.represent_scalar('tag:yaml.org,2002:python/name:'+name, '')

    def represent_module(self, data):
        return self.represent_scalar(
                'tag:yaml.org,2002:python/module:'+data.__name__, '')

    def represent_object(self, data):
        # We use __reduce__ API to save the data. data.__reduce__ returns
        # a tuple of length 2-5:
        #   (function, args, state, listitems, dictitems)

        # For reconstructing, we calls function(*args), then set its state,
        # listitems, and dictitems if they are not None.

        # A special case is when function.__name__ == '__newobj__'. In this
        # case we create the object with args[0].__new__(*args).

        # Another special case is when __reduce__ returns a string - we don't
        # support it.

        # We produce a !!python/object, !!python/object/new or
        # !!python/object/apply node.

        cls = type(data)
        if cls in copyreg.dispatch_table:
            reduce = copyreg.dispatch_table[cls](data)
        elif hasattr(data, '__reduce_ex__'):
            reduce = data.__reduce_ex__(2)
        elif hasattr(data, '__reduce__'):
            reduce = data.__reduce__()
        else:
            raise RepresenterError("cannot represent an object", data)
        reduce = (list(reduce)+[None]*5)[:5]
        function, args, state, listitems, dictitems = reduce
        args = list(args)
        if state is None:
            state = {}
        if listitems is not None:
            listitems = list(listitems)
        if dictitems is not None:
            dictitems = dict(dictitems)
        if function.__name__ == '__newobj__':
            function = args[0]
            args = args[1:]
            tag = 'tag:yaml.org,2002:python/object/new:'
            newobj = True
        else:
            tag = 'tag:yaml.org,2002:python/object/apply:'
            newobj = False
        function_name = '%s.%s' % (function.__module__, function.__name__)
        if not args and not listitems and not dictitems \
                and isinstance(state, dict) and newobj:
            return self.represent_mapping(
                    'tag:yaml.org,2002:python/object:'+function_name, state)
        if not listitems and not dictitems  \
                and isinstance(state, dict) and not state:
            return self.represent_sequence(tag+function_name, args)
        value = {}
        if args:
            value['args'] = args
        if state or not isinstance(state, dict):
            value['state'] = state
        if listitems:
            value['listitems'] = listitems
        if dictitems:
            value['dictitems'] = dictitems
        return self.represent_mapping(tag+function_name, value)

    def represent_ordered_dict(self, data):
        # Provide uniform representation across different Python versions.
        data_type = type(data)
        tag = 'tag:yaml.org,2002:python/object/apply:%s.%s' \
                % (data_type.__module__, data_type.__name__)
        items = [[key, value] for key, value in data.items()]
        return self.represent_sequence(tag, [items])

Representer.add_representer(complex,
        Representer.represent_complex)

Representer.add_representer(tuple,
        Representer.represent_tuple)

Representer.add_representer(type,
        Representer.represent_name)

Representer.add_representer(collections.OrderedDict,
        Representer.represent_ordered_dict)

Representer.add_representer(types.FunctionType,
        Representer.represent_name)

Representer.add_representer(types.BuiltinFunctionType,
        Representer.represent_name)

Representer.add_representer(types.ModuleType,
        Representer.represent_module)

Representer.add_multi_representer(object,
        Representer.represent_object)

